@@ -25,17 +25,20 @@ import java.util.Collection;
 import java.util.HashMap;
 import java.util.HashSet;
 import java.util.Iterator;
+import java.util.LinkedHashSet;
 import java.util.Map;
 
 import org.apache.maven.artifact.Artifact;
-import org.apache.maven.artifact.ArtifactUtils;
 import org.apache.maven.artifact.factory.ArtifactFactory;
 import org.apache.maven.artifact.installer.ArtifactInstaller;
 import org.apache.maven.artifact.repository.ArtifactRepository;
 import org.apache.maven.artifact.repository.ArtifactRepositoryFactory;
+import org.apache.maven.model.Model;
+import org.apache.maven.model.Parent;
 import org.apache.maven.plugin.AbstractMojo;
 import org.apache.maven.plugin.MojoExecutionException;
 import org.apache.maven.project.MavenProject;
+import org.codehaus.plexus.util.FileUtils;
 
 /**
  * Installs the project artifacts of the main build into the local repository as a preparation to run the sub projects.
@@ -48,7 +51,7 @@ import org.apache.maven.project.MavenProject;
  * @since 1.2
  * @author Paul Gier
  * @author Benjamin Bentmann
- * @version $Id: InstallMojo.java 698334 2008-09-23 20:16:00Z bentmann $
+ * @version $Id: InstallMojo.java 814226 2009-09-12 20:01:59Z bentmann $
  */
 public class InstallMojo
     extends AbstractMojo
@@ -109,6 +112,25 @@ public class InstallMojo
      */
     private Collection reactorProjects;
 
+    /**
+     * A flag used to disable the installation procedure. This is primarily intended for usage from the command line to
+     * occasionally adjust the build.
+     * 
+     * @parameter expression="${invoker.skip}" default-value="false"
+     * @since 1.4
+     */
+    private boolean skipInstallation;
+
+    /**
+     * The identifiers of already installed artifacts, used to avoid multiple installation of the same artifact.
+     */
+    private Collection installedArtifacts;
+
+    /**
+     * The identifiers of already copied artifacts, used to avoid multiple installation of the same artifact.
+     */
+    private Collection copiedArtifacts;
+
     /**
      * Performs this mojo's tasks.
      * 
@@ -117,11 +139,20 @@ public class InstallMojo
     public void execute()
         throws MojoExecutionException
     {
+        if ( skipInstallation )
+        {
+            getLog().info( "Skipping artifact installation per configuration." );
+            return;
+        }
+
         ArtifactRepository testRepository = createTestRepository();
 
-        installProjectArtifacts( project, testRepository );
-        installProjectParents( project, testRepository );
+        installedArtifacts = new HashSet();
+        copiedArtifacts = new HashSet();
+
         installProjectDependencies( project, reactorProjects, testRepository );
+        installProjectParents( project, testRepository );
+        installProjectArtifacts( project, testRepository );
     }
 
     /**
@@ -164,7 +195,9 @@ public class InstallMojo
     }
 
     /**
-     * Installs the specified artifact to the local repository.
+     * Installs the specified artifact to the local repository. Note: This method should only be used for artifacts that
+     * originate from the current (reactor) build. Artifacts that have been grabbed from the user's local repository
+     * should be installed to the test repository via {@link #copyArtifact(File, Artifact, ArtifactRepository)}.
      * 
      * @param file The file associated with the artifact, must not be <code>null</code>. This is in most cases the value
      *            of <code>artifact.getFile()</code> with the exception of the main artifact from a project with
@@ -187,7 +220,15 @@ public class InstallMojo
             {
                 throw new IllegalStateException( "Artifact is not fully assembled: " + file );
             }
-            installer.install( file, artifact, testRepository );
+
+            if ( installedArtifacts.add( artifact.getId() ) )
+            {
+                installer.install( file, artifact, testRepository );
+            }
+            else
+            {
+                getLog().debug( "Not re-installing " + artifact + ", " + file );
+            }
         }
         catch ( Exception e )
         {
@@ -195,6 +236,63 @@ public class InstallMojo
         }
     }
 
+    /**
+     * Installs the specified artifact to the local repository. This method serves basically the same purpose as
+     * {@link #installArtifact(File, Artifact, ArtifactRepository)} but is meant for artifacts that have been resolved
+     * from the user's local repository (and not the current build outputs). The subtle difference here is that
+     * artifacts from the repository have already undergone transformations and these manipulations should not be redone
+     * by the artifact installer. For this reason, this method performs plain copy operations to install the artifacts.
+     * 
+     * @param file The file associated with the artifact, must not be <code>null</code>.
+     * @param artifact The artifact to install, must not be <code>null</code>.
+     * @param testRepository The local repository to install the artifact to, must not be <code>null</code>.
+     * @throws MojoExecutionException If the artifact could not be installed (e.g. has no associated file).
+     */
+    private void copyArtifact( File file, Artifact artifact, ArtifactRepository testRepository )
+        throws MojoExecutionException
+    {
+        try
+        {
+            if ( file == null )
+            {
+                throw new IllegalStateException( "Artifact has no associated file: " + file );
+            }
+            if ( !file.isFile() )
+            {
+                throw new IllegalStateException( "Artifact is not fully assembled: " + file );
+            }
+
+            if ( copiedArtifacts.add( artifact.getId() ) )
+            {
+                File destination = new File( testRepository.getBasedir(), testRepository.pathOf( artifact ) );
+
+                getLog().debug( "Installing " + file + " to " + destination );
+
+                copyFileIfDifferent( file, destination );
+
+                MetadataUtils.createMetadata( destination, artifact );
+            }
+            else
+            {
+                getLog().debug( "Not re-installing " + artifact + ", " + file );
+            }
+        }
+        catch ( Exception e )
+        {
+            throw new MojoExecutionException( "Failed to stage artifact: " + artifact, e );
+        }
+    }
+
+    private void copyFileIfDifferent( File src, File dst )
+        throws IOException
+    {
+        if ( src.lastModified() != dst.lastModified() || src.length() != dst.length() )
+        {
+            FileUtils.copyFile( src, dst );
+            dst.setLastModified( src.lastModified() );
+        }
+    }
+
     /**
      * Installs the main artifact and any attached artifacts of the specified project to the local repository.
      * 
@@ -207,6 +305,7 @@ public class InstallMojo
     {
         try
         {
+            // Install POM (usually attached as metadata but that happens only as a side effect of the Install Plugin)
             installProjectPom( mvnProject, testRepository );
 
             // Install the main project artifact (if the project has one, e.g. has no "pom" packaging)
@@ -247,6 +346,7 @@ public class InstallMojo
             {
                 if ( parent.getFile() == null )
                 {
+                    copyParentPoms( parent.getGroupId(), parent.getArtifactId(), parent.getVersion(), testRepository );
                     break;
                 }
                 installProjectPom( parent, testRepository );
@@ -307,59 +407,80 @@ public class InstallMojo
         for ( Iterator it = reactorProjects.iterator(); it.hasNext(); )
         {
             MavenProject reactorProject = (MavenProject) it.next();
-            String id = reactorProject.getGroupId() + ':' + reactorProject.getArtifactId();
-            projects.put( id, reactorProject );
+
+            String projectId =
+                reactorProject.getGroupId() + ':' + reactorProject.getArtifactId() + ':' + reactorProject.getVersion();
+
+            projects.put( projectId, reactorProject );
         }
 
-        // collect transitive dependencies
-        Collection dependencies = new HashSet();
-        for ( Iterator it = mvnProject.getRuntimeArtifacts().iterator(); it.hasNext(); )
+        // group transitive dependencies (even those that don't contribute to the class path like POMs) ...
+        Collection artifacts = mvnProject.getArtifacts();
+        // ... into dependencies that were resolved from reactor projects ...
+        Collection dependencyProjects = new LinkedHashSet();
+        // ... and those that were resolved from the (local) repo
+        Collection dependencyArtifacts = new LinkedHashSet();
+        for ( Iterator it = artifacts.iterator(); it.hasNext(); )
         {
             Artifact artifact = (Artifact) it.next();
-            String id = ArtifactUtils.versionlessKey( artifact );
-            
-            dependencies.add( id );
+
+            // workaround for MNG-2961 to ensure the base version does not contain a timestamp
+            artifact.isSnapshot();
+
+            String projectId = artifact.getGroupId() + ':' + artifact.getArtifactId() + ':' + artifact.getBaseVersion();
+
+            if ( projects.containsKey( projectId ) )
+            {
+                dependencyProjects.add( projectId );
+            }
+            else
+            {
+                dependencyArtifacts.add( artifact );
+            }
         }
 
         // install dependencies
         try
         {
-            // install dependencies from reactor
-            for ( Iterator it = dependencies.iterator(); it.hasNext(); )
-            {
-                String id = (String) it.next();
-                MavenProject requiredProject = (MavenProject) projects.remove( id );
-                if ( requiredProject != null )
-                {
-                    it.remove();
-                    installProjectArtifacts( requiredProject, testRepository );
-                    installProjectParents( requiredProject, testRepository );
-                }
-            }
-
-            // install remaining dependencies from local repository
-            for ( Iterator it = mvnProject.getRuntimeArtifacts().iterator(); it.hasNext(); )
+            // copy dependencies that where resolved from the local repo
+            for ( Iterator it = dependencyArtifacts.iterator(); it.hasNext(); )
             {
                 Artifact artifact = (Artifact) it.next();
-                String id = ArtifactUtils.versionlessKey( artifact );
 
-                if ( dependencies.contains( id ) )
+                Artifact depArtifact =
+                    artifactFactory.createArtifactWithClassifier( artifact.getGroupId(), artifact.getArtifactId(),
+                                                                  artifact.getBaseVersion(), artifact.getType(),
+                                                                  artifact.getClassifier() );
+
+                File artifactFile = artifact.getFile();
+
+                Artifact pomArtifact =
+                    artifactFactory.createProjectArtifact( depArtifact.getGroupId(), depArtifact.getArtifactId(),
+                                                           depArtifact.getBaseVersion() );
+
+                File pomFile = new File( localRepository.getBasedir(), localRepository.pathOf( pomArtifact ) );
+
+                if ( pomFile.isFile() )
                 {
-                    File artifactFile = artifact.getFile();
-
-                    installArtifact( artifactFile, artifact, testRepository );
-
-                    Artifact pomArtifact =
-                        artifactFactory.createArtifact( artifact.getGroupId(), artifact.getArtifactId(),
-                                                        artifact.getVersion(), null, "pom" );
-
-                    File pomFile = new File( localRepository.getBasedir(), localRepository.pathOf( pomArtifact ) );
-
-                    if ( pomFile.exists() )
+                    if ( !pomArtifact.getId().equals( depArtifact.getId() ) )
                     {
-                        installArtifact( pomFile, pomArtifact, testRepository );
+                        copyArtifact( pomFile, pomArtifact, testRepository );
                     }
+                    copyParentPoms( pomFile, testRepository );
                 }
+
+                copyArtifact( artifactFile, depArtifact, testRepository );
+            }
+
+            // install dependencies that were resolved from the reactor
+            for ( Iterator it = dependencyProjects.iterator(); it.hasNext(); )
+            {
+                String projectId = (String) it.next();
+
+                MavenProject dependencyProject = (MavenProject) projects.get( projectId );
+
+                installProjectArtifacts( dependencyProject, testRepository );
+                installProjectParents( dependencyProject, testRepository );
             }
         }
         catch ( Exception e )
@@ -368,4 +489,50 @@ public class InstallMojo
         }
     }
 
+    /**
+     * Installs all parent POMs of the specified POM file that are available in the local repository.
+     * 
+     * @param pomFile The path to the POM file whose parents should be installed, must not be <code>null</code>.
+     * @param testRepository The local repository to install the POMs to, must not be <code>null</code>.
+     * @throws MojoExecutionException If any (existing) parent POM could not be installed.
+     */
+    private void copyParentPoms( File pomFile, ArtifactRepository testRepository )
+        throws MojoExecutionException
+    {
+        Model model = PomUtils.loadPom( pomFile );
+        Parent parent = model.getParent();
+        if ( parent != null )
+        {
+            copyParentPoms( parent.getGroupId(), parent.getArtifactId(), parent.getVersion(), testRepository );
+        }
+    }
+
+    /**
+     * Installs the specified POM and all its parent POMs to the local repository.
+     * 
+     * @param groupId The group id of the POM which should be installed, must not be <code>null</code>.
+     * @param artifactId The artifact id of the POM which should be installed, must not be <code>null</code>.
+     * @param version The version of the POM which should be installed, must not be <code>null</code>.
+     * @param testRepository The local repository to install the POMs to, must not be <code>null</code>.
+     * @throws MojoExecutionException If any (existing) parent POM could not be installed.
+     */
+    private void copyParentPoms( String groupId, String artifactId, String version, ArtifactRepository testRepository )
+        throws MojoExecutionException
+    {
+        Artifact pomArtifact = artifactFactory.createProjectArtifact( groupId, artifactId, version );
+
+        if ( installedArtifacts.contains( pomArtifact.getId() ) || copiedArtifacts.contains( pomArtifact.getId() ) )
+        {
+            getLog().debug( "Not re-installing " + pomArtifact );
+            return;
+        }
+
+        File pomFile = new File( localRepository.getBasedir(), localRepository.pathOf( pomArtifact ) );
+        if ( pomFile.isFile() )
+        {
+            copyArtifact( pomFile, pomArtifact, testRepository );
+            copyParentPoms( pomFile, testRepository );
+        }
+    }
+
 }
